/**********************************************************************************
 * NL nl.c
 * =================
 * versione    01a
 * autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
 *               LUGMan - Linux Users Group Mantova
 *               www.lugman.org
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.13.1
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * --------------------------------------------------------------------------------
 * Questo primo file mira semplicemente a creare un socket a livello kernel e a
 * stamparle il valore nel ring buffer. In questo modulo non ci sar la necessit
 * di creare un dispositivo e, per ora, neanche di creare l'interfaccia /proc.
 * La versoine 01a non tocca nulla nel modulo, ma implementa il programma in user-
 * space per inviare al kernel un messaggio.
 * --------------------------------------------------------------------------------
 * A livello implementativo comunque bisogna da subito tenere presente che la
 * funzione di creazione del socket  `netlink_kernel_create()' la quale prevede
 * come secondo parametro una funzione `input()' definita come:
 *   input (struct sock *sk, int len)
 * Questa funzione  una callback che DEVE essere definita, ma che in questa prima
 * versione non eseguir nulla.
 * Nel caso eseguisse qualche cosa, questa funzione verr chiamata automaticamente
 * ogni volta che un messaggio _arriva_ al netlink socket.
 * In realt il messaggio al netlink dovr essere nel protocollo descritto da una
 * macro (vedere linux/netlink.h) come NETLINK_{ROUTE,FIREWALL,ecc}. In questo
 * esempio verr definito un protocollo nuovo NETLINK_TEST come definito dal
 * documento 7356 di Linux Journal.
 *********************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
//--- NetLink ---
#include <linux/socket.h>  // gestione dei socket e strutture relative
#include <linux/netlink.h> // implementato in net/netlink/af_netlink.c (2.6.13.1)
#include <net/sock.h>      // implementa la struttura sock. senza questo  un casino referenziare i puntatori

MODULE_LICENSE("GPL");
MODULE_VERSION("0.1");

#define NETLINK_TEST 16

static struct sock *nl_sk = NULL;

/** _input  una callback che deve essere definita e passata alla funzione di
 * creazione del ntlink socket `netlink_kernel_create()'.
 * NOTA: se il manage del socket  abbastanza leggero, la funzione seguente pu
 * occuparsene direttamente, altrimenti  bene eseguire un wake_up_ in modo da
 * sbloccare l'esecuzione di un task sospeso appositamente.
 * */
void nl_input(struct sock *sk, int len)
{
   printk(KERN_ALERT "Messaggio ricevuto da user-space\n  sock= %d\n  len = %d\n",
	  (int)sk, len);
}


/** _init_module non fa altro che creare il socket NETLINK.
 * La creazione del socket a livello kernel si esegue con `netlink_kernel_create()'.
 * Non si usa netlink_create() la quale viene comunque richiamata da _kernel_create.
 * */
int nl_init_module(void)
{
   nl_sk = netlink_kernel_create(NETLINK_TEST,
				 nl_input);
   if(nl_sk)
     printk(KERN_INFO "Socket creato: %d\n", (int)nl_sk);
   else
     printk(KERN_ALERT "Socket non creato: %u\n", (unsigned)nl_sk);
   
   return 0;
}


/** _cleanup_module deve assolutamente eliminare il socket NETLINK.
 * NOTA: cosa succede sa il socket  aperto (x es in user-space) e scarico il 
 * modulo?
 * */
void nl_cleanup_module(void)
{
   printk(KERN_INFO "Close NETLINK socket %u...\n", (unsigned)nl_sk);
   sock_release(nl_sk->sk_socket);
}

module_init(nl_init_module);
module_exit(nl_cleanup_module);
