/**********************************************************************************
 * NL nl.c
 * =================
 * versione    06
 * autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
 *               LUGMan - Linux Users Group Mantova
 *               www.lugman.org
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.24.1 non retrocompatibile
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * --------------------------------------------------------------------------------
 * Fino a quanto fatto nella versione 05 il modulo poteva essere compilato solo se
 * la versione del kernel era minore o uguale alla 2.6.23. Dalla 2.6.24 infati il
 * Netlink ha subito cambiamenti strutturali. La versione 06  scritta in modo da
 * poter testare il driver su kernel a partire dal 2.6.24.
 * NON  mantenuta la retrocompatibilit per motivi di chiarezza.
 * 
 *   COSA CAMBIA:
 * - sock* netlink_kernel_create(): rispetto alla versione precedente introduce
 *  come primo parametro `struct net *net' i quale viene sempre passato come &init_net.
 * - void input(): cambia la funzione di input che oravule come parametro il buffer
 *  del socket (struct sk_buff *skb). Questo stravolge completamente le cose perch
 *  il sistema ora, quando riceve un mesaggio, estrae il buffer dal socket e lo
 *  passa a questa funzione. Il buffer quindi si pu leggenre, ma NON si pu pi
 *  rispedire o seindirizzare perch verr distrutto appena terminata questa call.
 *
 * Nelle versioni precedenti, l'indirizzo alla coda dei messaggi era tale per cui
 * si poteva rispedire il messaggio appena letto eventualmente ritoccandolo.
 * Ora non  pi cos. Ogni qual volta si vuole inviare un massaggio occorrer
 * creare e riempire il buffer (sk_buff*) il quale contiene l'header del netlink
 * seguito dal vero messaggio (payload).
 * NOTA: per creare il buffer si usa `alloc_skb()' che vuole come primo parametro
 * la dimensione dimensione di memoria del buffer+la lunghezza dell'header netlik
 * che dovrebbe essere 16 byte. Ad ogni modo ci si aiuta con la macro NLMSG_LENGTH.
 *
 * --------------------------------------------------------------------------------
 * IMPEMENTAZIONE:
 * l'implementazione  relativa principalmente alla callback input() associata al
 * buffer. L'accesso ai dati provenienti da user-space viene fatto con:
 * - sk_buff: definito in skbuff.h. Ha una serie infinita di campi anche a union
 *  e tra gli altri permette l'accesso diretto al buffer del socket tramite il
 *  campo data. Questo campo contiene anche l'header nlmsghdr.
 * - nlmsghdr: NETLINK Message Header definito in linux/netlink.h. Tra i vari campi
 *  continene anche nlmsg_pid ossia il pid del progesso mittente del messaggio.
 * Occorre decidere come "pilotare" l'invio del messaggio da parte del kernel. Ci
 * lo si pu fare in molti modi (proc, device driver, ecc), ma la via pi semplice
 * forse  inviare un messaggio con la richiesta di reinvio del messaggio.
 * L'invio di un messaggio unicast diretto a uno specifico processo da parte del 
 * kernel viene effettuato con la chiamata:
 *            netlink_unicast(struct sock, struct sk_buff, u32, int);
 * L'invio di un messaggio multicast viene eseguito con la chiamata:
 *           netlink_broadcast(struct sock, struct sk_buff, u32, 
 *                             u32 group, int allocation);
 * Come si nota la strtuttura necessaria  sempre sk_buff a cui si accede mediate
 * la macro NETLINK_CB() definita in linux/netlink.h
 *
 * 
 * 
 * CONTROLLA LA FRASE QUI SOTTO!!!!!
 * CONTROLLA LA FRASE QUI SOTTO!!!!!
 * 
 * E' stato poi introdotto il parametro `kernel_group' che permette di settare il
 * gruppo di appartenenza del kernel. Per fare questo baster modificare il
 * parametro tramite interfacci sysfs. si noti che questo gruppo non verr settato
 * fino a che il sistema non ricever un messaggio. La certezza del gruppo del
 * kernel  data da `actual_kernel_group'. Se i parametri sono uguali, il gruppo
 * sar settato correttamente.
 * --------------------------------------------------------------------------------
 * TEST:
 * Il test  condotto con nl_com per la versione 04 del modulo. Riferirisi alla
 * documentazione interna al file nl_mcom.c per vedere le varie modalit di test.
 * Anche il programma nl_com pu testare il modulo, ma  obsoleto.
 * --------------------------------------------------------------------------------
 * NOTA:
 *  il bug (o presunto tale) "KERNEL: assertion (!atomic_read(&sk->sk_rmem_alloc)) failed ..."
 * non si  pi riproposto.
 * --------------------------------------------------------------------------------
 * BUG:
 *********************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
//--- vrsione del kernel ---
#include <linux/version.h> // ATTENZIONE: nella versione 14 del kernel questo file non c'
//--- NetLink ---
#include <linux/socket.h>  // gestione dei socket e strutture relative
#include <linux/netlink.h> // implementato in net/netlink/af_netlink.c (2.6.13.1)
#include <net/sock.h>      // implementa la struttura sock. senza questo  un casino referenziare i puntatori
//--- nl.h ---
#include "nl.h"            // comprende macro e variabili necessarie al kernel quanto allo user space

#if LINUX_VERSION_CODE < 132622 // Kernel < di 1.6.24 ?
# error Kernel version too old!
#endif

MODULE_LICENSE("GPL");
MODULE_VERSION("0.6");
MODULE_AUTHOR("Calzo");
MODULE_DESCRIPTION("NL creates a NetLink socket and test the comunication between user and kernel space");

static unsigned kernel_group = 0;        // Gruppo pe la funzione netlink_kernel_create

module_param(kernel_group, uint, S_IRUGO|S_IWUSR);    // see LDD3 cap 2, pag 36 e il file linux/stat.h

MODULE_PARM_DESC(kernel_group, "Kernel Group for multicast comunications (field .dest_groups) ready after the nest step");
MODULE_PARM_DESC(use_skb_dequeue, "Use `skb_dequeue()' instead of `skb_recv_datagram()'");

static struct sock    *nl_sk  = NULL;

/** _change_str_case non fa altro che cambiare da minuscola a maiuscola e vice
 * versa la stringa passatagli.
 * */
void nl_change_str_case(char *str, int maxstr_len)
{
   int i;
   
   for(i=0; str[i] && i<maxstr_len; i++)
     {
	if(str[i]>='A' && str[i]<='Z')
	  str[i]+=0x20;
	else
	  if(str[i]>='a' && str[i]<='z')
	    str[i]-=0x20;
     }
}


/** _input  una callback che deve essere definita e passata alla funzione di
 * creazione del ntlink socket `netlink_kernel_create()'.
 * Questa funzione ha subito _strutturali_ cambiamenti dalla versione 2.6.24.
 * Ora infatti ha come primo e unico parametro il puntatore al buffer del socket
 * (sk_buff) il qualeverr poi liberato in automatico dal sistema.
 * Ci implica che non  pi necessario utilizzare le funzioni per il "dequeuing"
 * come `skb_recv_datagram()' o `skb_dequeue()'.
 * Di contro per il buffer ottenuto  da considerarsi in sola lettura o al massimo
 * modificabile solo all'interno di questa funzione. Se si vuole inviare un altro
 * messaggio occorre ricrearne il buffer.
 * 
 * La funzione quindi legge dallo user space i dati che il programma di test invia
 * per la generazione del socket (ossai leggiamo il payload).
 * A questo punto si procede creando un buffer sk_buff da inviare. La lunghezza del
 * buffer sar quella del payload+header di netlink (16 byte di slito).
 * Per creare il buffer si usa `allock_skb()'. Una volta ottenuto il buffer si
 * seleziona l'area dati e la si compila con i nuovi valori.
 * Si procede quindi all'invio che, come per le versioni precedenti,  determonato
 * dai valori contenuti nel payload ossia: se si vule mandare un messaggio unicast
 * non deve essere settato il multicast/broadcast e viceversa.
 * - multicast: questo messaggio si identifica semplicemnte settando il gruppo
 *  nel buffer (NETLINK_CB(skb).dst_group = gruppo) e ponendo lo stesso nella
 * funzione `netlink_broadcast()'. Si noti che passando un valore del pid diverso
 * da 0 a questa funzione, il processo identificato da quel pid non icever il
 * messaggio.
 * - unicast: questo  come prima. Basta passare il valore del PID
 *  
 * Si noti poi che  stata aggiunta a funzione `kfree_skb()' alla fine. In realt
 * dai commenti letti in net/core/datagram.c realtivamente a `skb_recv_datagram()'
 * la funzione migliore da utilizzare per liberare la memoria sarebbe
 * `skb_free_datagram()'. Per questa non fa altro che chiamare `kfree_skb()', ma
 * chiedendo 2 parametri! per questo mi sembra inutile chiamare `skb_free_datagram'
 * almeno fino al kernel 2.6.23.8 (controllato).
 * Ad ogni modo le funzioni di liberazione della memoria non sono state messe
 * perch hanno causato sempre crash gravi del modulo o peggio freeze dell'intero
 * sistema se poste nella funzione di clean_up.
 * --------------------------------------------------------------------------------
 *  NOTA: se il manage del socket  abbastanza leggero, la funzione seguente pu
 * occuparsene direttamente, altrimenti  bene eseguire un wake_up_ in modo da
 * sbloccare l'esecuzione di un task sospeso appositamente.
 *  NOTA: MSG_DONTWAIT permette di rendere la chiamata non-interruttiva. Credo che
 * questo provochi la perdita del messaggio se nessun socket  in ascolto, ma non
 * ne sono sicuro.
 * */

void nl_input(struct sk_buff *skb)
{
   struct sk_buff *buf = NULL;   // nuovo buffer che andremo a creare
   struct nlmsghdr *nlh;         // message header
   int len;
   struct nl_payload_t *payload_r, *payload_t; // payload ricevuto e da trasmettere
   
   printk(KERN_ALERT "Message received by user-space\n");
   
   // Comincio analizzando il buffer
   nlh = (struct nlmsghdr *)skb->data;                 // prelevo la struttura del mesaggio comprensiva di header
   payload_r = (struct nl_payload_t *)NLMSG_DATA(nlh); // prelevo il payload situato dopo l'header
   printk(KERN_ALERT "  sender pid: %d > %s\n",        // comunico il chi ha spedito il messaggio e cosa ha spedito
	  nlh->nlmsg_pid, payload_r->str);
   
   // Controllo che il messaggio non arrivi dal kernel per precauzione 
   if(!payload_r->multicast)
     if(NETLINK_CB(skb).pid == 0)
       {
	  printk(KERN_ALERT "Message sent by kernel: avoid processing\n");
	  return;
       }
      
   // Ora posso creare il nuovo buffer
   len = NLMSG_LENGTH(sizeof(struct nl_payload_t)); // calcolo la lunghezza totale del netlink
   buf = alloc_skb(len, GFP_KERNEL);   // creo il buffer
   printk("BUF = 0x%X\n", (int)buf);
   if(!buf)                            // se il buffer  nullo esco
     {
	printk(KERN_ERR "Buffer sk_buff created is null...\n");
	return; // e non devo liberare la memoria
     }

   // Ora mi sposto sull'area dati per selezionare il payload di trasmissione.
   // Questa chiamata (skb_put)  FONDAMENTALE. Senza questa il sistema non solo
   // non inizializza il buffer, ma l'invio del messaggio risulta inutile.
   nlh = (struct nlmsghdr *)skb_put(buf, len);          // prelevo la struttura del messaggio compreso l'header NETLINK
   // l'header va resettato o ricompilato in modo da istruire il processo in unser-
   // space. Chiaramente posso anche introdurre informazioni fasulle.
   memset(nlh, 0, len);
   nlh->nlmsg_pid = 0;  // sono info che il ricevente legger, ma sono fondamentalmente inutili in quanto posso anche falsificarle.
   nlh->nlmsg_len = len;  // gli passo anche la lunghezza del messaggio. Pu servire se si vuoleliberare la memoria tramite `skb_pull()'
   payload_t = (struct nl_payload_t *)NLMSG_DATA(nlh);  // prelevo il payload, ossia i dati effettivi
   
   // riempio il buffer
   strcpy(payload_t->str, payload_r->str); // copio il contenuto dell'uno nell'altro
   // rielaboro il payload tanto per far vedere che faccio qualche cosa
   nl_change_str_case(payload_t->str, MAX_PAYLOAD_CHAR);
   
   // L'istruzione che segue viene fatta sempre nel codice che manda informazioni
   // multicast (come in kobject_uevent.c). In realt per io non ho visto un vero
   // motivo per cui settarlo visto che qualsiasi numero metta sembra che il
   // comportamento non ne sia influenzato.
   NETLINK_CB(buf).dst_group = payload_r->multicast; // informazione apparentemente sperflua
   NETLINK_CB(buf).pid       = payload_r->pid;       // altra info superflua che non ho quasi mai visto usare

   payload_t->pid = 0;       // lo azzero per precauzione...  non si sa mai col multicast
   payload_t->multicast=0;   // idem
   
   if(payload_r->multicast)  /// controllo se mi  richiesto di spedire in Multicast...
     {
	// da questa versione ho notato che "payload_r->pid" implica che quel 
	// processo specifico NON ricever il messaggio
	netlink_broadcast(nl_sk, buf,           // socket a cui spedire il buffer "buf" 
			  payload_r->pid,       // PID di destinazione indicato dallo user-space
			  payload_r->multicast, // Multicast di destinazione indicato dallo user-space
			  GFP_ATOMIC);          // Modalit di allocazione (GFP_ATOMIC sarebbe pi efficiente)
	printk(KERN_ALERT "  MULTICAST Msg '%s' sent with group mask 0x%X - exclude pid %d\n",
	       payload_t->str, payload_r->multicast, payload_r->pid);
     } else                  /// ...ma se sono qui, spedisco in Unicast
     {
	netlink_unicast(nl_sk, buf, 
			payload_r->pid,  // PID di destinazione indicato dallo user-space  
			MSG_DONTWAIT);   // MSG_DONTWAIT=Chiamata non interruttiva
	printk(KERN_ALERT "  UNICAST Msg '%s' resent to %d process\n", 
	       payload_t->str,
	       payload_r->pid);
     }
   
__return:
// NOTA: a questo punto si potrebbe pensare di dover liberare la memoria con 
// `kfree_skb()', ma non serve pech la liberazione avviene da qualche altra parte
// perch se la eseguissi, il sistema andrebbe andrebbe in crash!!
     
#if 0    // il codice seuente rimuove i dati dal buffer, ma non serve di per s
   len = NLMSG_SPACE(nlh->nlmsg_len);
   if(len < buf->len)
     len = buf->len;
   skb_pull(buf, len);  // rimuovo i dati dall'inizio del buffer
#endif
}


/** _init_module non fa altro che creare il socket NETLINK.
 * La creazione del socket a livello kernel si esegue con `netlink_kernel_create()'.
 * Non si usa netlink_create() la quale viene comunque richiamata da _kernel_create.
 * */
int __init nl_init_module(void)
{
   nl_sk = netlink_kernel_create(&init_net,    // non so se va bene, am lo usano tutti
				 NETLINK_TEST, // TEST: valore arbitrario deciso da$
				 kernel_group,
				 nl_input,     // Funzione di gestione del socket
				 NULL, THIS_MODULE);
   if(nl_sk)
     printk(KERN_INFO "Socket created: %u (Group=%d)\n", (unsigned)nl_sk, kernel_group);
   else
     {
	printk(KERN_INFO "Socket not created: %u\n", (unsigned)nl_sk);
	return -ENOSPC;
     }
   
   return 0;
}


/** _cleanup_module deve assolutamente eliminare il socket NETLINK.
 * NOTA: cosa succede se il socket  aperto (x es in user-space) e scarico il 
 * modulo? ... uno strano messaggio
 * */
void nl_cleanup_module(void)
{
   printk(KERN_INFO "Close NETLINK socket %u...\n", (unsigned)nl_sk);
   sock_release(nl_sk->sk_socket);
}

module_init(nl_init_module);
module_exit(nl_cleanup_module);
