/**********************************************************************************
 * NL nl.c
 * =================
 * versione    06
 * autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
 *               LUGMan - Linux Users Group Mantova
 *               www.lugman.org
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.13.1
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * --------------------------------------------------------------------------------
 
      COMMENTARE
      COMMENTARE
 
 
 * Non cambia nulla rispetto alla 04 dal punto di vista del funzionamento, ma solo
 * viene implementato il codice per kernel maggiori del 2.6.14. Si noti che in
 * questa serie di kernel NON c' il file linux/version.h e le #define relative
 * al numero del kernel vengono crete dal Makefile. Roba da prendere qualche
 * "sviluppatore" e prenderlo a calci nei coglioni. "Ovviamente" se il kenrel  !=
 * dal 14 tutti i file esistono... x fortuna.
 * Dal kernel 2.6.14 sono cambiate in netlink.h:
 * - struct netlink_skb_parms: ridefinisce alcuni parametri e ne aggiunge altri
 *   relativo a SELinux. Quelli cancellati erano quelli che avevamo detto essere
 *   inutili.
 * - sock* netlink_kernel_create: aggiunge un "gruppo", mutex e module come
 *   parametri. Attualmente metto tutto a NULL e ho lo stesso funzionamento di
 *   prima.
 * --------------------------------------------------------------------------------
 * IMPEMENTAZIONE:
 * l'implementazione  relativa principalmente alla callback input() associata al
 * socket. L'accesso ai dati provenienti da user-space viene fatto con:
 * - sk_buff: definito in skbuff.h. Ha una serie infinita di campi anche a union
 *  e tra gli altri permette l'accesso diretto al buffer del socket tramite il
 *  campo data. Questo campo contiene anche l'header nlmsghdr.
 * - nlmsghdr: NETLINK Message Header definito in linux/netlink.h. Tra i vari campi
 *  continene anche nlmsg_pid ossia il pid del progesso mittente del messaggio.
 * Occorre decidere come "pilotare" l'invio del messaggio da parte del kernel. Ci
 * lo si pu fare in molti modi (proc, device driver, ecc), ma la via pi semplice
 * forse  inviare un messaggio con la richiesta di reinvio del messaggio.
 * L'invio di un messaggio unicast diretto a uno specifico processo da parte del 
 * kernel viene effettuato con la chiamata:
 *            netlink_unicast(struct sock, struct sk_buff, u32, int);
 * L'invio di un messaggio multicast viene eseguito con la chiamata:
 *           netlink_broadcast(struct sock, struct sk_buff, u32, 
 *                             u32 group, int allocation);
 * Come si nota la strtuttura necessaria  sempre sk_buff a cui si accede mediate
 * la macro NETLINK_CB() definita in linux/netlink.h
 * Per quanto riguarda i messaggi broadcast, il gruppo  un OR dei gruppi a cui il
 * messaggio  destinato, mantre allocation  il tipo di allicazione che il kernel
 * dovr eseguire per clonare il messaggio.
 * E' stato poi introdotto il parametro `kernel_group' che permette di settare il
 * gruppo di appartenenza del kernel. Per fare questo baster modificare il
 * parametro tramite interfacci sysfs. si noti che questo gruppo non verr settato
 * fino a che il sistema non ricever un messaggio. La certezza del gruppo del
 * kernel  data da `actual_kernel_group'. Se i parametri sono uguali, il gruppo
 * sar settato correttamente.
 * --------------------------------------------------------------------------------
 * TEST:
 * Il test  condotto con nl_com per la versione 04 del modulo. Riferirisi alla
 * documentazione interna al file nl_mcom.c per vedere le varie modalit di test.
 * Anche il programma nl_com pu testare il modulo, ma  obsoleto.
 * --------------------------------------------------------------------------------
 * BUG:
 *  KERNEL: assertion (!atomic_read(&sk->sk_rmem_alloc)) failed at net/netlink/af_netlink.c (122)
 * Non capisco a cosa sia dovuto. Da quanto letto nei commenti relativi alla
 * `netlink_kernel_create()' sembra che il buffer del socket vada liberato sempre
 * prima di uscire, ma il tentativo di liberare la memoria ha portato anche a gravi
 * freeze del sistema.
 * Comunque il problema non sembra grave e pare sia generato solo dopo aver tolto
 * il driver. Forse  una race (o simile) dovuta al fatto che la liberazione del 
 * socket elimina la struttura dellle "operazioni" `netlink_ops'
 *********************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
//--- vrsione del kernel ---
#include <linux/version.h> // ATTENZIONE: nella versione 14 del kernel questo file non c'
//--- NetLink ---
#include <linux/socket.h>  // gestione dei socket e strutture relative
#include <linux/netlink.h> // implementato in net/netlink/af_netlink.c (2.6.13.1)
#include <net/sock.h>      // implementa la struttura sock. senza questo  un casino referenziare i puntatori
//--- /proc ---
#include <linux/proc_fs.h>
//--- nl.h ---
#include "nl.h"            // comprende macro e variabili necessarie al kernel quanto allo user space

MODULE_LICENSE("GPL");
MODULE_VERSION("0.6");
MODULE_AUTHOR("Calzo");
MODULE_DESCRIPTION("NL creates a NetLink socket and test the comunication between user and kernel space");

#define NL_PROC_FILE "nl_send_msg"

struct proc_dir_entry *proc_nl = NULL;
static unsigned kernel_group = 0;        // Gruppi di appartenenza in attesa di definizione
static unsigned actual_kernel_group = 0; // Gruppo attuale di appartenenza reale in questo momento
static unsigned use_skb_dequeue = 0;     // Usa skb_dequeue() al posto di skb_recv_datagram()

module_param(kernel_group, uint, S_IRUGO|S_IWUSR);    // see LDD3 cap 2, pag 36 e il file linux/stat.h
module_param(actual_kernel_group, uint, S_IRUGO);     // see LDD3 cap 2, pag 36 e il file linux/stat.h
module_param(use_skb_dequeue, uint, S_IRUSR|S_IWUSR); // see LDD3 cap 2, pag 36 e il file linux/stat.h

MODULE_PARM_DESC(kernel_group, "Kernel Group for multicast comunications (fielad .dest_groups) ready to the nest step");
MODULE_PARM_DESC(actual_kernel_group, "Actual Kernel Group for multicast comunications (fielad .dest_groups");
MODULE_PARM_DESC(use_skb_dequeue, "Use `skb_dequeue()' instead of `skb_recv_datagram()'");

static struct sock    *nl_sk  = NULL;

/** _change_str_case non fa altro che cambiare da minuscola a maiuscola e vice
 * versa la stringa passatagli.
 * */
void nl_change_str_case(char *str, int maxstr_len)
{
   int i;
   
   for(i=0; str[i] && i<maxstr_len; i++)
     {
	if(str[i]>='A' && str[i]<='Z')
	  str[i]+=0x20;
	else
	  if(str[i]>='a' && str[i]<='z')
	    str[i]-=0x20;
     }
}


/** _input  una callback che deve essere definita e passata alla funzione di
 * creazione del ntlink socket `netlink_kernel_create()'.
 * Questa funzione invia dal kernel allo user-space un messaggio a fronte di un
 * altro messaggio proveniente dallo user-space. Questo serve solo per triggherare
 * l'evento di invio da parte del kernel ed  una scelta arbitraria.
 * In funzione da quanto richiesto dallo user-space il messaggio pu essere 
 * unicast o multicast. Alcune note relativamente a queste due modalit:
 * - unicast: viene lanciata con MSG_DONTWAIT in modo che non sia interrttiva. La
 *   macro  definita in linux/socket.h. Consultarlo per le altre modalit.
 * - multicast: la funzione invia a un gruppo di processi il messaggio.
 * Viene poi azzerato nlh->nlmsg_pid = 0; questo permette al processo di capire
 * che il mittenete  il kernel e cos potremo vedere se due processi possono
 * usare NetLink in user-space dove questo valore sar diverso da 0.
 * 
 * Si noti poi che  stata aggiunta a funzione `kfree_skb()' alla fine. In realt
 * dai commenti letti in net/core/datagram.c realtivamente a `skb_recv_datagram()'
 * la funzione migliore da utilizzare per liberare la memoria sarebbe
 * `skb_free_datagram()'. Per questa non fa altro che chiamare `kfree_skb()', ma
 * chiedendo 2 parametri! per questo mi sembra inutile chiamare `skb_free_datagram'
 * almeno fino al kernel 2.6.23.8 (controllato).
 * Ad ogni modo le funzioni di liberazione della memoria non sono state messe
 * perch hanno causato sempre crash gravi del modulo o peggio freeze dell'intero
 * sistema se poste nella funzione di clean_up.
 * --------------------------------------------------------------------------------
 *  NOTA: se il manage del socket  abbastanza leggero, la funzione seguente pu
 * occuparsene direttamente, altrimenti  bene eseguire un wake_up_ in modo da
 * sbloccare l'esecuzione di un task sospeso appositamente.
 *  NOTA: MSG_DONTWAIT permette di rendere la chiamata non-interruttiva. Credo che
 * questo provochi la perdita del messaggio se nessun socket  in ascolto, ma non
 * ne sono sicuro.
 * */
void nl_input(struct sk_buff *skb)
{
   struct nlmsghdr *nlh_s;
   struct nl_payload_t *payload_s;
   
   printk(KERN_INFO "Message received by user-space\n");
   
   nlh_s = (struct nlmsghdr *)skb->data; 
   if(!nlh_s)
     {
	printk("ERROR: skn->data == NULL!\n");
	return;
     }
   
   payload_s = (struct nl_payload_t *)NLMSG_DATA(nlh_s);  // prelevo il payload, ossia i dati effettivi
   printk(KERN_ALERT "  sender pid: %d > %s\n", 
	  nlh_s->nlmsg_pid, payload_s->str);
}




int nl_read_proc(char *buf, char **start, off_t offset,
		 int count, int *eof, void *data)
{
   struct sk_buff *skb=NULL;
   size_t len;
 
   printk(KERN_INFO "NL: accesso a PROC\n");
   
   if(!nl_sk)
     {
	printk(KERN_INFO "ERROR: Netlink Socket non valido\n");
	return -1;
     }
   
   /* allocate message with the maximum possible size */
   //len =  NLMSG_SPACE(sizeof(struct nl_payload_t));
   len=NLMSG_LENGTH(sizeof(struct nl_payload_t));
   skb = alloc_skb(len, GFP_KERNEL);
   if (skb) 
     {
 	struct nlmsghdr *nlh;
	struct nl_payload_t *buffer;
	
 /* con questo codice dimostro che io ho allocato un'area di memoria nella quale 
  * mi devo muovere. Tramite _put credo di potermi muovere internamente, mentre
  * con _push e _pop aggiungo o sottraggo specifici messaggi. Inoltre stampando gli
  * indirizzi vedo che skb->data  = al risultato di skb_put, ma  0x10 (16) byte
  * minore di dove in realt risiede il payload. Praticamente il primo skb_put
  *  come se creasse l'header, come scritto anche in kobject_uevebt.c. 
  * NOTA: skb_push mi manda in crash il sistema */
	printk(" data   = %X\n", (int)skb->data);
	buffer = (struct nl_payload_t *)skb_put(skb, len);
 /* ALERT: se il codice appena qui sopra non viene settato credo che la struttura
  * del socket non contenga i dati perch non viene aggiornato len della struttura.
  * Per capire queste cose devi studiare la funzione inline skb_put e simili */
	nlh = (struct nlmsghdr *)skb->data;
	printk(" buffer = %X\n data   = %X\n", (int)buffer, (int)skb->data);
	
	//buffer = (struct nl_payload_t *)NLMSG_DATA(nlh); //skb_put(skb, len);
	buffer = (struct nl_payload_t *)NLMSG_DATA(nlh);
	//buffer = (struct nl_payload_t *)skb_put(skb, 0);
	printk(" buffer = %X\n", (int)buffer);

	buffer->pid=2;
	buffer->multicast=1;
	strcpy(buffer->str, "StringA\0");
	printk(" str = %s\n", buffer->str);
	
	NETLINK_CB(skb).dst_group = 1;
	netlink_broadcast(nl_sk, skb, 0, 1, GFP_KERNEL);
     } else
     printk("WARNING: skb NULL!\n");
      
   return 0;
}





/** _init_module non fa altro che creare il socket NETLINK.
 * La creazione del socket a livello kernel si esegue con `netlink_kernel_create()'.
 * Non si usa netlink_create() la quale viene comunque richiamata da _kernel_create.
 * */
int __init nl_init_module(void)
{
#if LINUX_VERSION_CODE >= 132632      // kernel >= 2.6.24
   printk("Kernel >= 2.6.24\n");
   nl_sk = netlink_kernel_create(&init_net,    // non so se va bene, am lo usano tutti
				 NETLINK_TEST, // TEST: valore arbitrario deciso da$
				 1,
				 nl_input,     // Funzione di gestione del socket
				 NULL, THIS_MODULE);
#elif LINUX_VERSION_CODE >= 132622    // kernel >= 2.6.14...
   printk("Kernel >= 2.6.14\n");
   nl_sk = netlink_kernel_create(NETLINK_TEST, // TEST: valore arbitrario deciso da$
				 0, //NETLINK_TEST,
				 nl_input,     // Funzione di gestione del socket
				 NULL, THIS_MODULE);
#else      // se siamo nel kernel < di 2.6.14
   printk("Kernel < 2.6.14\n");
   nl_sk = netlink_kernel_create(NETLINK_TEST, // TEST: valore arbitrario deciso da noi
				 nl_input);    // Funzione di gestione del socket
#endif
   if(nl_sk)
     printk(KERN_INFO "Socket created: %u\n", 
	    (unsigned)nl_sk);
   else
     {
	printk(KERN_INFO "Socket not created: %u\n", (unsigned)nl_sk);
	return -1;
     }

   proc_nl = create_proc_read_entry(NL_PROC_FILE,
				    0, NULL,
				    nl_read_proc, NULL);
   if(proc_nl==NULL)
      printk(KERN_ALERT "WARNING: wrong init /proc filesystem!\n");
   
   return 0;
}


/** _cleanup_module deve assolutamente eliminare il socket NETLINK.
 * NOTA: cosa succede se il socket  aperto (x es in user-space) e scarico il 
 * modulo? ... uno strano messaggio
 * */
void nl_cleanup_module(void)
{
   printk(KERN_INFO "Close NETLINK socket %u...\n", (unsigned)nl_sk);
   sock_release(nl_sk->sk_socket);
   if(proc_nl)
     remove_proc_entry(NL_PROC_FILE,    // file da cancellare
		       NULL);           // parent dir
}

module_init(nl_init_module);
module_exit(nl_cleanup_module);
